;  __   ___   ___  _  _   ___   _  
; |  \ |   | |    | \/ | |   | | \ |
; |__/ |---| |    |    | |---| |  \|
; |    |   | |___ |    | |   | |   |
;
; written by blackbird
; copyright by namco

	processor f8
	include "ves.h"

;--------------------;
; Register Reference ;
;--------------------;

; The registers in the f8 are used as both RAM and registers for
; this game. The reference is as follows:
;
; r0-r31	general-purpose registers
; r33-r34	level data address
; r35		pac-man x coordinate
; r36		pac-man y coordinate
; r37		pac-man direction
; r38		animation index

;---------------;
; Program Entry ;
;---------------;

	org	$800

cartridgeStart:
	CARTRIDGE_START

cartridgeEntry:
	lis	0						; init the h/w
	outs	1
	outs	4
	outs	5
	outs	0
                
	lisu	4						; r32 = complement flag
	lisl	0
	lr	S, A
                
	li	$c6						; set to three color, grey background
	lr	3, A						; clear screen to grey
	pi	clrscrn						;

;-------------------;
; Main Program Loop ;
;-------------------;
				
mainloop:

displayLevel0:
	; load in the address for level 0
	lisu	4
	lisl	1
	li	>level0						; load upper address of level 0
	lr	I, A						; store into r33, and advance ISAR
	li	<level0						; load lower address of level 0
	lr	S, A						; store into r34	

	li	bkg
	lr	1, A
	li	blue
	lr	2, A
	pi	drawMap						; draw the map

	pi	drawSidebar					; and the sidebar

setupPacman:
	; load coordinates
	lisu	4
	lisl	3
	li	38
	lr	S, A						; load the pac-man x coordinate
	lisl	4
	li	34
	lr	S, A						; load the pac-man y coordinate

	; set the direction
	lisl	5
	li	%00000001
	lr	S, A

	; load the animation index
	lisl	6
	li	0
	lr	S, A

playLoop:
.clearPacman:
	; clear pac-man
	li	bkg
	lr	1, A
	; load x coordinate
	lisu	4
	lisl	3
	lr	A, S
	lr	2, A
	; load y coordinate
	lisl	4
	lr	A, S
	lr	3, A
	; set sprite address
	li	>clearSprite
	lr	Qu, A
	li	<clearSprite
	lr	Ql, A
	; draw the sprite
	pi	drawSprite

.checkController:
	; handle input and pacman's direction
	pi	pacmanControls

.checkDirection:
	; make sure current direction is still valid
	pi	checkIntersection
	ci	1
	bnz	.checkDirectionEnd				; not in an intersection

	; clear direction if necessary
	lisu	4
	lisl	5
	lr	A, S
	ns	0
	lr	S, A	
.checkDirectionEnd:

.movePacman:
	; now we move pac-man based on the direction
	; which is set in r34

	lisl	5
	lr	A, S
	lr	0, A
	; up
	ni	%00001000
	bnz	.movePacmanUp
	; down
	lr	A, 0
	ni	%00000100
	bnz	.movePacmanDown
	; left
	lr	A, 0
	ni	%00000010
	bnz	.movePacmanLeft
	; right
	lr	A, 0
	ni	%00000001
	bnz	.movePacmanRight
	; none
	br	.movePacmanEnd

.movePacmanUp:
	; subtract y by 1
	lisl	4
	lr	A, S
	ai	%11111111					; subtract 1
	lr	S, A
	br	.movePacmanEnd
.movePacmanRight:
	; add x by 1
	lisl	3
	lr	A, S
	ai	1						; add 1
	lr	S, A
	br	.movePacmanEnd
.movePacmanDown:
	; add y by 1
	lisl	4
	lr	A, S
	ai	1						; add 1
	lr	S, A
	br	.movePacmanEnd
.movePacmanLeft:
	; subtract x by 1
	lisl	3
	lr	A, S
	ai	%11111111					; subtract 1
	lr	S, A
	br	.movePacmanEnd
.movePacmanEnd:

.drawPacman:
	; get sprites' address
	li	>pacmanSprites
	lr	Qu, A
	li	<pacmanSprites
	lr	Ql, A
	; transfer it into the DC
	lr	DC, Q
	; find the sprite offset
	lisu	4
	lisl	6
	lr	A, S						; load the sprite index
	inc							; make sure we at least hit 0
	lr	0, A
	li	5						; load A with an increment of 5
.drawPacmanGetAddress:
	ds	0
	bz	.drawPacmanSaveAddress
	adc							; add 8 for every tile to the DC
	br	.drawPacmanGetAddress
.drawPacmanSaveAddress:
	; transfer back the address
	lr	Q, Dc

	; draw the sprite
	li	green
	lr	1, A
	; load x coordinate
	lisu	4
	lisl	3
	lr	A, S
	lr	2, A
	; load y coordinate
	lisl	4
	lr	A, S
	lr	3, A
	; draw the sprite
	pi	drawSprite

.updatePacmanFrame:
	lisu	4
	lisl	6
	lr	A, S						; load the sprite index
	inc							; increase it
	ci	4						; if we weren't on frame 3
	bnz	.updatePacmanFrameEnd				; then don't reset it
	li	0
.updatePacmanFrameEnd:
	lr	S, A

	; delay for a little bit
	lis	10
	lr	5, A
	pi	delay

	jmp	playLoop

endPlayLoop:
	jmp	mainloop

;----------------------;
; Controller Functions ;
;----------------------;

; spools for controller input,
; returns with controller input in A

readController:
	; see one of the hand controllers is moved
	lis	0
	outs	0
	lis	0
	outs	4
	ins	4
	com							; un-invert port data
	ni	$cf						; mask off twists, since we don't use them
	bnz	.readControllerEnd
	outs	1
	ins	1
	com
	ni	$cf
.readControllerEnd:
	pop

; manages controller input and how to
; move pac-man

pacmanControls:
	; save return address
	lr	K, P
	pi	pushk

	; get controller input
	pi	readController
	; and save it
	lr	2, A
.pacmanControlsSetDirection:
	; check to see if we're on an intersection
	pi	checkIntersection

	; branch if false
	ci	1
	bnz	.pacmanControlsNoIntersection

.pacmanControlsInIntersection:
	; load the boundaries
	lr	A, 0

	; compare and set direction
	br	.pacmanControlsSetEnd

.pacmanControlsNoIntersection:
	; if you're not in an intersection,
	; you can only go the opposite way
	; you're going right now

	; set ISAR for direction
	lisu	4
	lisl	5

	; load temp value in r0
	lr	A, S
	lr	0, A

	; up
	lr	A, 0
	ni	%00001000
	bnz	.pacmanControlsCheckUp
	; down
	lr	A, 0
	ni	%00000100
	bnz	.pacmanControlsCheckDown
	; left
	lr	A, 0
	ni	%00000010
	bnz	.pacmanControlsCheckLeft
	; right
	lr	A, 0
	ni	%00000001
	bnz	.pacmanControlsCheckRight
	; none
	clr
	br	.pacmanControlsSetEnd

.pacmanControlsCheckUp:
	li	%00000100
	br	.pacmanControlsSetEnd
.pacmanControlsCheckDown:
	li	%00001000
	br	.pacmanControlsSetEnd
.pacmanControlsCheckLeft:
	li	%00000001
	br	.pacmanControlsSetEnd
.pacmanControlsCheckRight:
	li	%00000010

.pacmanControlsSetEnd:
	; see if the direction is allowed
	ns	2
	; invalid direction, forget about it
	bz	.pacmanControlsEnd

	; set ISAR for direction
	lisu	4
	lisl	5
	; load the direction in
	lr	S, A

.pacmanControlsEnd:
	pi	popk
	pk							; return from the subroutine
	

; checks to see if pacman is in an intersection
; if so, it returns with A set to 1, else it
; returns with A set to 0. Also, if pacman is, it 
; returns with the intersection boundaries in r0

; modifies: ISAR, r0, r1

checkIntersection:
	lisu	4
	lisl	3
	lr	A, S
	ni	%00000111
	ci	2
	bnz	.checkIntersectionFalse
	lisl	4
	lr	A, S
	ni	%00000111
	ci	2
	bz	.checkIntersectionTrue	
.checkIntersectionFalse:
	lis	0
	pop
.checkIntersectionTrue:
	; load the map address
	lisu	4
	lisl	1						; r33
	lr	A, I						; get the high byte
	lr	Qu, A
	lr	A, S						; and the low byte
	lr	Ql, A
	; transfer it into the DC
	lr	DC, Q

	; add y offset to DC
	lisl	4
	lr	A, D						; r36
	sr	1
	sr	1
	sr	1
	inc							; make sure we hit 0
	lr	0, A
	lis	10
.checkIntersectionYLoop:
	ds	0
	bz	.checkIntersectionYLoopEnd
	adc
	br	.checkIntersectionYLoop
.checkIntersectionYLoopEnd:

	; add x offset to DC
	lr	A, S						; r35
	sr	1
	sr	1
	sr	1
	adc

	; get the tile number
	lm
	; now we find this tile's boundaries

	inc							; make sure we hit 0
	lr	0, A
	; get the tile address
	li	>tiles
	lr	Qu, A
	li	<tiles
	lr	Ql, A
	; load the 16-bit address into the DCI
	lr	DC, Q
	lis	3						; three bytes for each tile
.checkIntersectionAddressLoop:
	ds	0
	bz	.checkIntersectionAddressLoopEnd
	adc
	br	.checkIntersectionAddressLoop
.checkIntersectionAddressLoopEnd:
	lm
	lm
	lm
	lr	0, A						; save the boundary for the tile

.checkIntersectionEnd:
	lis	1
	pop

;-------------------;
; Drawing Functions ;
;-------------------;

; plot out a single point on the screen
; uses three registers as "arguments"
; r1 = color
; r2 = x (to screen) (0-101)
; r3 = y (to screen) (0-57)

plot:
	; set the color using r1
	; colors: $40-red, $80=blue, $0=green, $c0=bkg
	lr	A, 1
	outs	1

	; set the column using r2
	lr	A, 2
	ai	4
	com
	outs	4

	; set the row using r3
	lr	A, 3
	ai	4
	com
	outs	5

	; transfer data to the screen memory
	li	$60
	outs	0
	li	$50
	outs	0

	; delay until it's fully updated
	lis	6
.plotDelay:	
	ai	$ff
	bnz	.plotDelay

	pop							; return from the subroutine

; draw an 8x8 tile from a data pointer
; r1 = color 1
; r2 = color 2
; r3 = x (to screen)
; r4 = y (to screen)
; r5 = tile number

drawTile:
	; registers reference:
	; r1 = reserved (plot color)
	; r2 = reserved (plot x)
	; r3 = reserved (plot y)
	; r4 = color 1
	; r5 = color 2
	; r6 = x
	; r7 = y
	; r8 = loop1 (row)
	; r9 = loop2 (column)
	; r10 = bitmask
	; r11 = graphics byte

	; save the return address
	lr	K, P
	pi	pushk

	; get the tile address
	li	>tiles
	lr	Qu, A
	li	<tiles
	lr	Ql, A
	; load the 16-bit address into the DCI
	lr	DC, Q
	; add the offset
	lr	A, 5
	inc							; make sure we hit 0
	lr	0, A
	lis	3						; three bytes for each tile
.drawTileAddressLoop:
	ds	0
	bz	.drawTileAddressLoopEnd
	adc
	br	.drawTileAddressLoop
.drawTileAddressLoopEnd:
	; got the tile data, now get the graphics address
	lm
	lr	Qu, A
	lm
	lr	Ql, A
	lr	DC, Q

	; rearrange the registers
	lr	A, 4
	lr	7, A
	lr	A, 3
	lr	6, A
	lr	A, 2
	lr	5, A
	lr	A, 1
	lr	4, A

	; start row loop
	li	8
	lr	8, A
.drawTileRow:
	; setup the bit mask
	li	%10000000					; we only want the left-most pixel
	lr	10, A						; store it in r10
	; and get the graphics byte
	lm							; load the pixel and increase the DC
	lr	11, A						; store it in r11

	; start column loop
	li	8
	lr	9, A
.drawTileColumn:

	; find out the x address
	lr	A, 9						; load the new offset
	com							; and subtract it from 8
	inc							; via 2's complement
	ai	8						; subtract A from 8, ignore carry
	as	6						; add the tile's real x offset
	lr	2, A						; save it for the subroutine call
	; and the y address
	lr	A, 8						; load the new offset
	com							; and subtract it from 8
	inc							; via 2's complement
	ai	8						; subtract A from 8, ignore carry
	as	7						; add the tile's real y offset
	lr	3, A						; save it for the subroutine call

	; initially load color 2
	lr	A, 5
	lr	1, A
	; see if the pixel is on or off
	lr	A, 11						; load the byte
	ns	10						; mask it with the bitmask
	bnz	.drawTilePixel					; if the pixel's off, use color 1
	; conditionally load color 1
	lr	A, 4
	lr	1, A

.drawTilePixel:
	pi	plot

.drawTileRollBitmask:
	; roll the bitmask
	lr	A, 10
	sr	1
	lr	10, A

.drawTileCheckLoops:
	; check column loop
	ds	9
	bnz	.drawTileColumn

	; check row loop
	ds	8
	bnz	.drawTileRow

	pi	popk
	pk							; return from the subroutine

; draw a map from a map pointer
; r1 = color 1
; r2 = color 2
; Q = address of map

drawMap:
	; this function uses several reserve registers
	; because the functions called need the immediate ones
	; r17 (21o) = x counter
	; r18 (22o) = y counter
	; r19 (23o) = color 1
	; r20 (24o) = color 2
	; r21 (25o) = address offset

	; load the return address into memory
	lr	K, P
	pi	pushk

	; clear address offset
	lisu	2
	lisl	5
	clr
	lr	S, A

	; backup the colors
	; color 1
	lisl	3
	lr	A, 1
	lr	I, A						; r23
	; color 2
	lr	A, 2
	lr	S, A						; r24

	lisl	1						; 
	lis	7						; load r19 with 7 rows of tiles
	lr	S, A						;
.drawMapRow:

	lisl	2						;
	lis	10						; load r20 with 10 tiles per row
	lr	S, A						;
.drawMapColumn:

	; this is the start of drawing an individual tile

	; note of reference to drawTile:
	; r1 = color 1
	; r2 = color 2
	; r3 = x (to screen)
	; r4 = y (to screen)
	; r5 = tile number

	lisl	2						; load register r20
	lr	A, S
	lr	0, A						; set r0 as our row counter
	li	0						; clear A to the first pixel
.drawMapGetX:
	ds	0
	bz	.drawMapSaveX
	ai	8
	br	.drawMapGetX

.drawMapSaveX:
	com							; we want to display the tiles from
	inc							; top to bottom, so we inverse
	ai	72						; both coordinates
	lr	3, A						; store the x coordinate in r3

	lisl	1						; load register r19
	lr	A, S
	lr	0, A						; set r0 as our column counter
	li	0						; clear A to zero
.drawMapGetY:
	ds	0
	bz	.drawMapSaveY
	ai	8
	br	.drawMapGetY

.drawMapSaveY:
	com							; we want to display the tiles from
	inc							; top to bottom, so we inverse
	ai	48						; both coordinates
	lr	4, A						; store the y coordinate in r4

.drawMapSetupRegisters:

	; this block of code gets the next tile number
	; and increases the data offset
	; this needs to be optimized...

	; load the map address
	lisu	4
	lisl	1						; r33
	lr	A, I						; get the high byte
	lr	Qu, A
	lr	A, S						; and the low byte
	lr	Ql, A
	; transfer it into the DC
	lr	DC, Q
	; load the map offset
	lisu	2
	lisl	5
	lr	A, S
	; and add it to the DC
	adc
	; increase the map offset
	inc
	lr	S, A
	; get the next tile byte and save it
	lm
	lr	5, A

	; restore the colors
	; color 1
	lisl	3
	lr	A, I						; r21
	lr	1, A						; into r1
	; color 2
	lr	A, S						; r22
	lr	2, A						; into r2

	; finally, call the tile function
	pi	drawTile

.drawMapCheckLoops:
	; check column loop
	lisu	2
	lisl	2
	ds	D						; r18
	bnz	.drawMapColumn

	; check row loop
	ds	S						; r17
	bnz	.drawMapRow

.drawMapReturn:
	; return to the program flow
	pi	popk
	pk							; return from the subroutine

; draw an 8x8 sprite from a data pointer
; r1 = color
; r2 = x (to screen)
; r3 = y (to screen)
; Q = 16-bit address

drawSprite:
	; registers reference:
	; r1 = reserved (plot color)
	; r2 = reserved (plot x)
	; r3 = reserved (plot y)
	; r4 = color
	; r6 = x
	; r7 = y
	; r8 = loop1 (row)
	; r9 = loop2 (column)
	; r10 = bitmask
	; r11 = graphics byte
	; Q = tile address
	; K = return address

	; save the return address
	lr	K, P
	; load the 16-bit address into the DCI
	lr	DC, Q

	; rearrange the registers
	lr	A, 3
	lr	7, A
	lr	A, 2
	lr	6, A
	lr	A, 1
	lr	4, A

	; start row loop
	li	5
	lr	8, A
.drawSpriteRow:
	; setup the bit mask
	li	%10000000					; we only want the left-most pixel
	lr	10, A						; store it in r10
	; and get the graphics byte
	lm							; load the pixel and increase the DC
	lr	11, A						; store it in r11

	; start column loop
	li	5
	lr	9, A
.drawSpriteColumn:

	; find out the x address
	lr	A, 9						; load the new offset
	com							; and subtract it from 5
	inc							; via 2's complement
	ai	5						; subtract A from 5, ignore carry
	as	6						; add the tile's real x offset
	lr	2, A						; save it for the subroutine call
	; and the y address
	lr	A, 8						; load the new offset
	com							; and subtract it from 5
	inc							; via 2's complement
	ai	5						; subtract A from 5, ignore carry
	as	7						; add the tile's real y offset
	lr	3, A						; save it for the subroutine call

	; see if the pixel is on or off
	lr	A, 11						; load the byte
	ns	10						; mask it with the bitmask
	bz	.drawSpriteRollBitmask				; if the pixel's off, don't draw anything
	; conditionally load color 1
	lr	A, 4
	lr	1, A

.drawSpritePixel:
	pi	plot

.drawSpriteRollBitmask:
	; roll the bitmask
	lr	A, 10
	sr	1
	lr	10, A

.drawSpriteCheckLoops:
	; check column loop
	ds	9
	bnz	.drawSpriteColumn

	; check row loop
	ds	8
	bnz	.drawSpriteRow

	pk							; return from the subroutine

; draws the right sidebar rectangle

drawSidebar:
	; save return address
	lr	K, P

	; set sidebar color
	li	blue
	lr	1, A

	; start row loop
	li	57
	lr	3, A
.drawSidebarRow:
	li	101
	lr	2, A
.drawSidebarColumn:
	pi	plot

.drawSidebarCheck:
	ds	2
	lr	A, 2
	xi	%01010000
	bnz	.drawSidebarColumn
	ds	3
	bp	.drawSidebarRow
.drawSiderbarEnd:
	pk

;-----------;
; Tile Data ;
;-----------;

; format:
; two bytes for the address of the tile
; one byte for the boundaries of the tile (%0000 + up + down + left + right)

tiles:
	; these are regular tiles
	; which can be used in any map

	; tile 0
	.word	tilePic_0
	.byte	%00001111

	; tile 1
	.word	tilePic_1
	.byte	%00000101

	; tile 2
	.word	tilePic_2
	.byte	%00000111

	; tile 3
	.word	tilePic_3
	.byte	%00000011

	; tile 4
	.word	tilePic_4
	.byte	%00000110

	; tile 5
	.word	tilePic_5
	.byte	%00001001

	; tile 6
	.word	tilePic_6
	.byte	%00001110

	; tile 7
	.word	tilePic_7
	.byte	%00001011

	; tile 8
	.word	tilePic_8
	.byte	%00001101

	; tile 9
	.word	tilePic_9
	.byte	%00001100

	; tile 10
	.word	tilePic_10
	.byte	%00001111

	; tile 11
	.word	tilePic_11
	.byte	%00001010

	; these bytes are for the ghost
	; area: blank, and you can only
	; move left or right

	; tile 12
	.word	tilePic_0
	.byte	%00000001

	; tile 13
	.word	tilePic_0
	.byte	%00000010

;---------------;
; Tile Graphics ;
;---------------;

tilePic_0:
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000

tilePic_1:
	.byte	%00000000
	.byte	%00111111
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000

tilePic_2:
	.byte	%00000000
	.byte	%11111111
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%10000000

tilePic_3:
	.byte	%00000000
	.byte	%11111111
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%11111111

tilePic_4:
	.byte	%00000000
	.byte	%11111110
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%10000001

tilePic_5:
	.byte	%01000001
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%00111111

tilePic_6:
	.byte	%01000001
	.byte	%10000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%10000001

tilePic_7:
	.byte	%01000001
	.byte	%10000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%11111111

tilePic_8:
	.byte	%01000001
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000
	.byte	%01000000

tilePic_9:
	.byte	%01000001
	.byte	%01000001
	.byte	%01000001
	.byte	%01000001
	.byte	%01000001
	.byte	%01000001
	.byte	%01000001
	.byte	%01000001

tilePic_10:
	.byte	%01000001
	.byte	%10000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%00000000
	.byte	%10000000

tilePic_11:
	.byte	%01000001
	.byte	%10000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%00000001
	.byte	%11111110

;------------;
; Level Data ;
;------------;

level0:
	; level map
	.byte	1,  2,  2,  3,  4,  1,  3,  2,  2,  4
	.byte	5,  6,  5,  2,  7,  7,  2, 11,  8, 11
	.byte	0,  9,  1, 10,  3,  3, 10,  4,  9,  0
	.byte	3, 10,  6,  9,  0,  0,  9,  8, 10,  3
	.byte	0,  9,  8, 10,  3,  3, 10,  6,  9,  0
	.byte	1,  7, 11,  5,  4,  1, 11,  5,  7,  4
	.byte	5,  3,  3,  3,  7,  7,  3,  3,  3, 11

	; pellet map

;----------------;
; Pacman Sprites ;
;----------------;

clearSprite:
	; 5x5 block to clear
	; a drawn sprite
	.byte	%11111000
	.byte	%11111000
	.byte	%11111000
	.byte	%11111000
	.byte	%11111000

pacmanSprites:
	; sprite 0
	.byte	%01110000
	.byte	%11000000
	.byte	%11000000
	.byte	%11000000
	.byte	%01110000

	; sprite 1
	.byte	%01110000
	.byte	%11110000
	.byte	%11100000
	.byte	%11110000
	.byte	%01110000

	; sprite 2
	.byte	%01110000
	.byte	%11111000
	.byte	%11111000
	.byte	%11111000
	.byte	%01110000

	; sprite 3
	.byte	%01110000
	.byte	%11110000
	.byte	%11100000
	.byte	%11110000
	.byte	%01110000

;-----------;
; Signature ;
;-----------;

	org $ff0

signature:
	.byte	"Blackbird 2004"
